#The Rethinking library has a few requirements
#install.packages("rstan")
#install.packages("StanHeaders")
#install.packages(c("coda","mvtnorm","devtools","loo"))
#library(devtools)
#devtools::install_github("rmcelreath/rethinking")

library(rethinking)


#First, let's just fit a normal distribution

#Here's the model
f <- alist(
  y ~ dnorm( mu , sigma ),
  mu ~ dnorm( 0 , 10 ),
  sigma ~ dcauchy( 0 , 1 )
)

#Now we fit it
fit <- map( 
  f , 
  data=list(y=c(-1,1)) , 
  start=list(mu=0,sigma=1)
)

#now let's fit it using STAN
fit.stan <- map2stan( 
  f , 
  data=list(y=c(-1,1)) , 
  start=list(mu=0,sigma=1)
)

#Great - now let's build a dataset
#where x1 -> y1 -> y2
set.seed(31415)
ndf <- data.frame(x1=runif(100,0,100))
ndf$y1 <- rnorm(100, ndf$x1*2, 40)
ndf$y2 <- rnorm(100, ndf$y1*2, 80)

#Now, let's build a model
nf <- alist(
  
  #derived values
  #y1Pred <- yhat1,
  y2Pred <- b2*b1*x1,
  resid_y1 <- y1-yhat1,
  resid_y2 <- y2-yhat2,
  
  #The observed variables & error
  y1 ~ dnorm(yhat1, sigma1),
  y2 ~ dnorm(yhat2, sigma2),
  
  #underlying prediction relationships
  yhat2 <- b2*y1,
  yhat1 <-  b1*x1,
  

  
  #priors for slopes
  b1 ~ dnorm( 0 , 1000 ),
  b2 ~ dnorm( 0 , 1000 ),
  
  #priors for variances
  sigma1 ~ dcauchy( 0 , 1 ),
  sigma2 ~ dcauchy( 0 , 1 )
  
)

#Fit the SEM with STAN
n.fit <- map2stan( 
  nf , 
  data=ndf , 
  start=list(b1=1, b2=1, sigma1=10, sigma2=10),
  chains=4 , cores=2 , iter=2000 , warmup=1000
)

#Evaluate parameters
summary(n.fit)

#look at the trace plot
plot(n.fit)

#examine the correlation between parameters
pairs(n.fit)

#diagnostics and residual plots?


#Evaluate residuals
pred_and_resid <- link(n.fit, data=ndf)

pred_and_resid_data <- as.data.frame(lapply(pred_and_resid, colMeans))

pred_and_resid_data <- cbind(ndf, pred_and_resid_data)

plot(y1 ~ yhat1, data=pred_and_resid_data)
plot(y2 ~ yhat2, data=pred_and_resid_data)

plot(resid_y1 ~ yhat1, data=pred_and_resid_data)
plot(resid_y2 ~ yhat2, data=pred_and_resid_data)

qqnorm(pred_and_resid_data$resid_y1)
qqline(pred_and_resid_data$resid_y1, col="red")

qqnorm(pred_and_resid_data$resid_y2)
qqline(pred_and_resid_data$resid_y2, col="red")



#This will do prediction, for yhat values
link(n.fit, data=list(x1=3)) -> a
lapply(a, mean)

sims <- sim(n.fit, data=list(x1=3))


#Compare to lavaan fit
library(lavaan)

mod <- '
 y1 ~ x1
 y2 ~ y1
'

summary(sem(mod, data=ndf))

#####
###Factor analytic model - THIS DOES NOT WORK
#####

#In a Latent Variable model, we treat LV's as missing data
#that are filled in with NAs
set.seed(31415)
latent_df <- data.frame(LV=rnorm(100,5,10))#, X_Latent=NA)
latent_df <- within(latent_df, {
  x1 <- rnorm(100, LV*1, 60)
  x2 <- rnorm(100, LV*3.5, 80)
  x3 <- rnorm(100, LV*1.4, 50)
  idx<-1:100
  
})

#Now the model. Note that we specify that some latent X exists
#and has a distribution
latent_mod <- alist(
    X_Latent ~ dnorm(mu_N, sigma_N),

    x1 ~ dnorm(xhat1, sigma1),
    xhat1 <-  1*X_Latent,
    
    x2 ~ dnorm(xhat2, sigma2),
    xhat2 <-  b2*X_Latent,
    
    x3 ~ dnorm(xhat3, sigma3),
    xhat3 <-  b3*X_Latent,
  
    mu_N ~ dnorm( 0 , 1000 ),
    b3 ~ dnorm( 0 , 1000 ),
    b2 ~ dnorm( 0 , 1000 ),
  
    sigma1 ~ dcauchy( 0 , 1 ),
    sigma2 ~ dcauchy( 0 , 1 ),
    sigma3 ~ dcauchy( 0 , 1 ),
    sigma_N ~ dcauchy( 0 , 1 )
  
)

#Fit the model - 
latent_fit <- map2stan( 
  latent_mod , 
  data=latent_df , 
  start=list(b2=1, b3=1, sigma1=10, sigma2=10, sigma3=10, sigma_N=10),
  warmup=500 , iter=5500
)
